#!/usr/bin/env python

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = []
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = ""
__email__ = ""
__status__ = "Development"
