#!/usr/bin/env python
"""cluster: provides tools for clustering
"""

__all__ = ['goodness_of_fit', 'metric_scaling', 'procrustes', 'UPGMA', 'nmds',
           'approximate_mds']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Catherine Lozuopone", "Rob Knight", "Peter Maxwell",
    "Justin Kuczynski"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Catherine Lozupone"
__email__ = "lozupone@colorado.edu"
__status__ = "Production"
