#!/usr/bin/env python
__all__ = ["test_best_likelihood",
           "test_bootstrap",
           "test_likelihood_function",
           "test_motifchange",
           "test_parameter_controller",
           "test_scale_rules",
           "test_simulation",
           "test_substitution_model",
           "test_coevolution",
           "test_models"]

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Rob Knight", "Peter Maxwell","Greg Caporaso"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Rob Knight"
__email__ = "rob@spot.colorado.edu"
__status__ = "Production"
