#!/usr/bin/env python
"""apps: provides support libraries for controlling applications (local or web).
"""

__all__ = ['blast',
           'carnac',
           'cd_hit',
           'clustalw',
           'cmfinder',
           'comrna',
           'consan',
           'contrafold',
           'cove',
           'dialign',
           'dynalign',
           'fasttree',
           'fasttree_v1'
           'foldalign',
           'gctmpca',
           'ilm',
           'knetfold',
           'mfold',
           'muscle',
           'msms',
           'nupack',
           'parameters',
           'pfold',
           'pknotsrg',
           'raxml',
           'rdp_classifier',
           'rnaalifold',
           'rnaforester',
           'rnashapes',
           'rnaview',
           'sfold',
           'unafold',
           'util',
           'vienna_package']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Jeremy Widmann", "Catherine Lozuopone", "Gavin Huttley",
               "Rob Knight", "Zongzhi Liu", "Sandra Smit", "Micah Hamady",
               "Greg Caporaso", "Mike Robeson", "Daniel McDonald", 
               "Marcin Cieslik"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Rob Knight"
__email__ = "rob@spot.colorado.edu"
__status__ = "Production"

"""Need to add:
    tcoffee     alignment
    meme        motif finding
    phylip      phylogeny
    mrbayes     phylogeny
    paml        phylogeny
    
    other packages? web tools?

"""
