#!/usr/bin/env python
"""db: provides support libraries for database retrieval.
"""

__all__ = ['ncbi', 'util', 'rfam', 'pdb', 'ensembl']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Rob Knight", "Mike Robeson", "Zongzhi Liu", "Gavin Huttley",
               "Hua Ying"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Rob Knight"
__email__ = "rob@spot.colorado.edu"
__status__ = "Production"

"""Need to add:

    rfam            RNA families database
    go              Gene Ontology
    kegg            KEGG metabolic pathways and related file formats
    pfam            Protein families database
    pdb             Protein Data Bank
    ndb             Nucleotide Data Bank ('Atlas' files)
"""
