#!/usr/bin/env python
__all__ = ['test_fit_function',
           'test_geometry',
           'test_matrix',
           'test_matrix_logarithm',
           'test_optimisers',
           'test_stats',
           'test_unifrac',
           'test_spatial']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Rob Knight", "Peter Maxwell", "Catherine Lozupone",
               "Gavin Huttley", "Sandra Smit", "Marcin Cieslik", "Antonio Gonzalez Pena"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Rob Knight"
__email__ = "rob@spot.colorado.edu"
__status__ = "Production"
