#!/usr/bin/env python
__all__ = ['array', 'checkpointing', 'datatypes', 'dict2d', 'misc', 'modules',
           'organizer', 'parallel', 'table', 'transform', 'unit_test',
           'warning','recode_alignment']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Gavin Huttley", "Rob Knight", "Sandra Smit",
                    "Peter Maxwell", "Amanda Birmingham", "Zongzhi Liu",
                    "Andrew Butterfield", "Daniel McDonald",
                    "Greg Caporaso"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
