#!/usr/bin/env python

__all__ = ["best_likelihood", "bootstrap", "likelihood_calculation",
           "likelihood_function", "likelihood_tree", "models",
           "pairwise_distance", "parameter_controller", "predicate",
           "simulate", "substitution_calculation", "substitution_model",
           "coevolution"]

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Gavin Huttley", "Peter Maxwell", "Andrew Butterfield",
                    "Rob Knight", "Matthrew Wakefield", "Brett Easton", 
                    "Edward Lang","Greg Caporaso"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
