#!/usr/bin/env python

__all__ = ['test_alignment',
           'test_alphabet',
           'test_annotation',
           'test_bitvector',
           'test_core_standalone',
           'test_entity',
           'test_genetic_code',
           'test_info',
           'test_location',
           'test_maps',
           'test_moltype',
           'test_profile',
           'test_sequence',
           'test_tree',
           'test_tree2',
           'test_usage']

__author__ = ""
__copyright__ = "Copyright 2007-2012, The Cogent Project"
__credits__ = ["Catherine Lozupone", "Peter Maxwell", "Rob Knight",
                    "Gavin Huttley", "Jeremy Widmann", "Greg Caporaso",
                    "Sandra Smit", "Justin Kuczynski", "Marcin Cieslik"]
__license__ = "GPL"
__version__ = "1.5.3"
__maintainer__ = "Rob Knight"
__email__ = "rob@spot.colorado.edu"
__status__ = "Production"
